%skeleton "lalr1.cc" /* -*- C++ -*- */
%require "2.1a"
%defines
%define "parser_class_name" "ClausesParser"

%{
#include "Term.h"
#include "Constant.h"
#include "Function.h"
#include "Variable.h"
#include "datatypes.h"
#include "Clause.h"
#include "ConcreteClause.h"
#include "VariableNameGenerator.h"
#include "ClauseContext.h"

#include <string>
#include <vector>
#include <set>
class ClausesDriver;
%}

// The parsing context.
%parse-param { ClausesDriver& driver }
%lex-param { ClausesDriver& driver }

%locations
%initial-action
{
// Initialize the initial location.
@$.begin.filename = @$.end.filename = &driver.file;
};

%debug
%error-verbose

// Symbols.
%union
{
NS_PROOFLIB::Term *term;
NS_PROOFLIB::TermList *term_list;
std::string *sval;
Clause *clause;
std::vector<NS_PROOFLIB::Term*> *term_vector;
std::set<Clause*> *clause_set;
};

%{
# include "ClausesDriver.h"
%}

%token END 0 "end of file"
%token OBRACE "("
%token EBRACE ")"
%token OCBRACE "{"
%token ECBRACE "}"
%token COMMA ","
%token <sval> UPCASE "upper_case_id"
%token <sval> LOWCASE "lower_case_id"
%token DOT "."
%token IMPLIED ":-"
%token EQ "="

%type <term> term "term"
%type <term> function "function"
%type <term> constant "constant"
%type <term> variable "variable"
%type <term> eqp "eqp"
%type <term_list> term_list "term_list"
%type <clause> clause "clause"
%type <term_vector> term_vector "term_vector"
%type <clause_set> clause_set "clause_set"

%printer { debug_stream () << *$$; } "upper_case_id"
%printer { debug_stream () << *$$; } "lower_case_id"
%printer { debug_stream () << $$->getStringRep(); } "function"
%printer { debug_stream () << $$->getStringRep(); } "constant"
%printer { debug_stream () << $$->getStringRep(); } "variable"
%printer { debug_stream () << $$->getStringRep(); } "term"

%destructor { delete $$; } "upper_case_id"
%destructor { delete $$; } "lower_case_id"
%destructor { delete $$; } "clause"
%destructor { delete $$; } "term_vector"
%destructor { delete $$; } "clause_set"
%destructor { $$->detach(); } "function"
%destructor { $$->detach(); } "constant"
%destructor { $$->detach(); } "variable"

%destructor { $$->detach(); } "term"

%%

%start end_term;

end_term: clause_set { driver.result = $1; };

clause_set: clause_set clause
{
	$1->insert($2);
}
| /* Nothing */
{
	$$ = new std::set<Clause*>();
}
;

clause: OCBRACE OCBRACE term_vector ECBRACE COMMA OCBRACE term_vector ECBRACE ECBRACE
{
	Clause *cl = new ConcreteClause($3, $7, driver.vng);
    cl->setClauseContext(new ClauseContext());
    $$ = cl;
};

clause: term_vector IMPLIED term_vector DOT
{
	Clause *cl = new ConcreteClause($1, $3, driver.vng);
    cl->setClauseContext(new ClauseContext());
    $$ = cl;
};

term_vector: term_vector COMMA term
{
	$1->push_back($3);
}
| term_vector term
{
	$1->push_back($2);
}
| /* Nothing */
{
	$$ = new std::vector<NS_PROOFLIB::Term*>();
}
;

term: function {}
| constant {}
| variable {}
| eqp {}
;

term_list: term_list COMMA term
{
	$1->addRight($3);
}
| term
{
	NS_PROOFLIB::TermList *l = new NS_PROOFLIB::TermList();
	l->addRight($1);
	$$ = l;
}
;

function: LOWCASE OBRACE term_list EBRACE
{
	$$ = driver.pdb->newFunction( $1->c_str(), *($3) ) ;
};

eqp: term EQ term
{
    NS_PROOFLIB::TermList *l = new NS_PROOFLIB::TermList();
	l->addRight($1);
	l->addRight($3);
    $$ = driver.pdb->newFunction( "=", *l ) ;
};

constant: LOWCASE {$$ = driver.pdb->newConstant( $1->c_str()) ;};

variable: UPCASE {$$ = driver.pdb->newVariable( $1->c_str()) ;};

%%

void
yy::ClausesParser::error (const yy::ClausesParser::location_type& l,
const std::string& m)
{
	driver.error (l, m);
}
