#ifndef CLAUSE_H_
#define CLAUSE_H_

#include "Variable.h"
#include "Function.h"
#include "Constant.h"
#include "Term.h"
#include "VariableNameGenerator.h"
#include "DataProofNode.h"
#include "ClauseContext.h"
#include "ATPException.h"
#include "ClauseInfo.h"

#include <string>
#include <map>
#include <vector>
#include <set>

class Clause : public DataProofNode
{
	ClauseContext *context;
	VariableNameGenerator *vng;
    std::vector<NS_PROOFLIB::Term *> pos;
    std::vector<NS_PROOFLIB::Term *> neg;
    ClauseInfo *info;
public:
	Clause(std::vector<NS_PROOFLIB::Term *> pos, std::vector<NS_PROOFLIB::Term *> neg, VariableNameGenerator*);
	virtual ~Clause();
	/* Variant should change the variable names in the context completely so there would be no
	 * x -> x_1 but x_1 -> ? as we merge the contexts in the unifier and the original variables
	 * may appear in both.
	 * also binds it to its parent clause
	 */
	Clause *getVariant();
	ClauseInfo *getClauseInfo();
	VariableNameGenerator *getVariableNameGenerator();
    std::vector<NS_PROOFLIB::Term*> &getPositiveLiterals();
    std::vector<NS_PROOFLIB::Term*> &getNegativeLiterals();
    ClauseContext *getClauseContext();
    void setClauseContext(ClauseContext *cc);
    std::string toString();
    bool operator<(const Clause &t) const;   // equal
    // includes variables that are mapped to closed terms also
    std::set<NS_PROOFLIB::Variable*> getAllVariables();
    // return a random, with unique constants, ground clause
    // also keep track on all these constants as they should be deleted
    // when the clause is deleted (unlike normal clauses which are never deleted)
    Clause *getGroundInstance();
    // this removes bindings which do not appear in the clause any more
    // this can cause an occur check in unification when it is no longer the case
    //void clean();
    virtual void print(NodePrint&);
private:
	NS_PROOFLIB::Term *mapVariants(NS_PROOFLIB::Term *t, std::map<NS_PROOFLIB::Term*, NS_PROOFLIB::Term*> &variants);
    void addVariables(NS_PROOFLIB::Term *told, std::set<NS_PROOFLIB::Variable*> &st) const;
    std::string termToString(NS_PROOFLIB::Term*);
    void recGround(NS_PROOFLIB::Term*, ClauseContext*, int&);
};

// clauses with less literals are smaller
class ClausePointerCmp
{
public:
	bool operator()(const Clause*, const Clause*) const;
};

// clauses with less literals are smaller
class ClausePairCmp
{
public:
	bool operator()(const std::pair<Clause*,unsigned int>, const std::pair<Clause*,unsigned int>) const;
};

class Variant : public Clause
{
    public:
    Variant(std::vector<NS_PROOFLIB::Term *> pos, std::vector<NS_PROOFLIB::Term *> neg, VariableNameGenerator*);
    virtual ~Variant();
    virtual void print(NodePrint&);
};

/*
for printing factors but as C++ supports polymorphism so poorly, although the factors are implemented in another package, an interface
must be created here
*/
class Factor : public Clause
{
    private:
    bool isPos;
    public:
    Factor(bool isPositive, std::vector<NS_PROOFLIB::Term *> pos, std::vector<NS_PROOFLIB::Term *> neg, VariableNameGenerator*);
    virtual ~Factor();
    virtual void print(NodePrint&);
    bool isPositive();
};

#endif /*CLAUSE_H_*/
