#include "PositiveRefinement.h"

using namespace std;
using namespace NS_PROOFLIB;

PositiveRefinement::PositiveRefinement()
{
	useHistory = false;
}

PositiveRefinement::~PositiveRefinement()
{
}

void PositiveRefinement::initialize(InputSet *cs)
{
	set<Clause*> *clauses = cs->getClauseSet();
	for (set<Clause*>::const_iterator it = clauses->begin(); it != clauses->end(); ++it)
	{
		Clause *clause = *it;
		insertClause(clause);
	}
	
	// initialization of indices
	firstInd = -1;
	secondInd = 0;
}

void PositiveRefinement::insertClause(Clause* c)
{
	if (all.size() >= MAX_SIZE)
	{
		throw new RefinementException("The size of the clause set has exceeded the max value which was set in the refinement (for dealing with the history), you can raise this value or find a better hashing algorithm");
	}
	if (isPositive(c))
	{
		pos.push_back(true);
	}
	else
	{
		pos.push_back(false);
	}
	all.push_back(c);
}

bool PositiveRefinement::isPositive(Clause *c)
{
	return (c->getPositiveLiterals().size() > 0 && c->getNegativeLiterals().size() == 0);
}

void PositiveRefinement::finalize()
{
	
}

// if the last positive already did the last clause in the vector
bool PositiveRefinement::isEmpty() const
{
	return false; // emptiness is determined by getNextResolvingSequence returning NULL
}

ResolvingSequence *PositiveRefinement::getNextResolvingSequence()
{	
	setNextIndices();
	// set exhausted
	if (firstInd == -1 || firstInd >= secondInd) return NULL;
	ResolvingSequence *ret = new ResolvingSequence(new SimpleStrategy());
	vector<Clause*> rsl;
	rsl.push_back(all[firstInd]);
	rsl.push_back(all[secondInd]);
	ret->initialize(rsl);
	
	if (useHistory) history.insert((MAX_SIZE * firstInd) + secondInd);
	
	return ret;
}

void PositiveRefinement::setNextIndices()
{
	firstInd++;
	if (firstInd == secondInd)
	{
		if (secondInd < all.size()-1)
		{
			secondInd++;
			firstInd = 0;
		}
		else
		{
			// set exhausted
			return;
		}
	}
	
	// ensure one is positive
	if(!pos[firstInd] && !pos[secondInd])
	{
		setNextIndices();
		return;
	}
	
	// ensure not in history
	if (useHistory && history.find((MAX_SIZE * firstInd) + secondInd) != history.end())
	{
		setNextIndices();
		return;
	}
}

map<Clause*, int> PositiveRefinement::getIndicesMap()
{
	if (!useHistory) useHistory = true;
		
	map<Clause*, int> ret;
	
	typedef vector<Clause*>::const_iterator IT;
	
	int i = 0;
	for (IT it = all.begin(); it != all.end(); ++it)
	{
		ret[*it] = i;
		i++;
	}
	
	return ret;
}

string PositiveRefinement::getIndicesMessage()
{
	return "Please choose two clauses to process: ";
}

ResolvingSequence *PositiveRefinement::getResolvingSequenceFromIndices(vector<int> l)
{
	unsigned int index1 = l[0];
	unsigned int index2 = l[1];
	
	// index1 must always be smaller
	if (index1 > index2)
	{
		int tmp = index1;
		index1 = index2;
		index2 = tmp;
	}
	
	vector<Clause*> rsl;
	rsl.push_back(all[index1]);
	rsl.push_back(all[index2]);
	ResolvingSequence *rs = new ResolvingSequence(new SimpleStrategy()); 
	rs->initialize(rsl);
	
	// here it is interactive so we always use history
	history.insert((MAX_SIZE * index1) + index2);
	
	return rs;
}

void PositiveRefinement::setEventPerformed(InputSetEvent ise)
{
	switch (ise.getType())
	{
	case InputSetEvent::CLAUSE_ADDED:
		insertClause(ise.getClause());
	    break;
	case InputSetEvent::CLAUSE_DELETED:
		throw new ATPException("deletion not implemented yet in positive refinement");
	case InputSetEvent::CLAUSE_REFUSED:
		//nothing to do
		break;
	}
}
