#include "PromptUserCommand.h"
#include "DeductionEngine.h"
#include "SimpleRefinement.h"
#include "UnificationResolver.h"
#include "InputSet.h"
#include "Refinement.h"
#include "ResolvingUnit.h"
#include "SimpleVariableNameGenerator.h"
#include "SimpleInsertPostProcessor.h"

#include "LatexNodePrint.h"

#include "ClauseStrategyRuling.h"
#include "ClauseContext.h"
#include "ProofProcess.h"

#include "datatypes.h"
#include "Term.h"
#include "Constant.h"
#include "Variable.h"
#include "Function.h"

#include <typeinfo>
#include <iostream>
#include <fstream>
#include <set>
#include <vector>

using namespace std;
using namespace NS_PROOFLIB;

PromptUserCommand::PromptUserCommand()
{
}

PromptUserCommand::~PromptUserCommand()
{
}

void PromptUserCommand::execute()
{
	try
	{
		InputSet *is = env->getClausesInput()->parseFromFile(env->getInputFilename());

		// initialize post processors
		typedef std::set<ResolventPostProcessor*, PostProcessorPriorityCmp>::const_iterator LCI;
		set<ResolventPostProcessor*, PostProcessorPriorityCmp> postProcessors = env->getPostProcessors();
		for (LCI i = postProcessors.begin(); i != postProcessors.end(); ++i)
		{
			ResolventPostProcessor *rpp = *i;
			rpp->init(is);
		}

		process = new ProofProcess(env);
		process->initialize(is);

		Resolvent * res = process->execute();

		cout<<"Total Elapsed Time: "<<process->getElapsedTime()<<endl;
		cout<<"Machine Elapsed Time: "<<process->getMachineTime()<<endl;

		NodePrint *printer = env->getNodePrint();
		if (res != NULL && res->getType() == Resolvent::EMPTY)
		{
			cout<<"Successfully refuted input set of clauses"<<endl;
			cout<<"The list of unifiers is:"<<endl;


			printer->initialize();
			printer->beginTree();
			res->print(*printer);
			printer->endTree();
			printer->finalize();
		}
		// print all end resolvents trees
		else
		{
			cout<<"Could not refute input set of clauses. Generating full resolvents trees (if any)"<<endl;
			set<Clause*> *cs = process->getInputSet()->getClauseSet();

            printer->initialize();

			for (set<Clause*>::const_iterator it = cs->begin(); it != cs->end(); ++it)
			{
				Clause *c = *it;
				//is end resolvent
				if (typeid(*c) == typeid(Resolvent) && (dynamic_cast<ConcreteClause*>(c))->getResolvedClauses().empty())
				{
					printer->beginTree();
					c->print(*printer);
					printer->endTree();
				}
			}
					printer->finalize();
		}
    }
    catch (ATPException *a)
    {
        cout<<"Exception in main.cc: "<<a->getMessage()<<endl;
    }
    catch (...)
    {
    	delete process;
    }
}

vector<int> PromptUserCommand::printInfo(map<Clause*, int> m)
{
	// also order the clauses according to indices so they given indices will stay the same (mapped indices wwill be equal to the actual order
	// of the clauses on the screen. This way it will be easier to locate new claues (at the end of the list)

	set<pair<Clause*, int>, printComp> s(m.begin(), m.end());

	vector<int> mappedIndices(m.size());
	int loc = 0;
	for (set<pair<Clause*, int>, printComp>::const_iterator it = s.begin(); it != s.end(); ++it)
	{
		pair<Clause*, int> p = *it;
		mappedIndices[loc] = p.second;
		cout<<"("<<loc<<") "<<p.first->toString()<<"\t";
		loc++;
	}
	return mappedIndices;
}

UserReply *PromptUserCommand::retrieveInfo(InputSet *is, map<Clause*, int> m, string s, vector<int> &mappedIndices)
{
	// check if refinement supports macro steps
	string macroStep;
	bool macroEnabled = false;
	if (typeid(*(env->getManagedRefinement())) == typeid(MacroStepRefinement))
	{
		macroEnabled = true;
		macroStep = ", m(<list of clauses separated by ,>)(<target clause>) for macro deduction";
	}

	cout<<endl<<"Enter tX for automated X steps"<<macroStep<<", sX for displaying all clauses containing X or "<<s;
	int a = -1,b;
	int ato = -1;
	string searchTerm;
	cin>>a;
	// automatic mode or search mode
	if (!cin)
	{
		cin.clear();
		string s;
		cin>>s;
		if (s.substr(0,1) == "t")
		{
			// automatic mode
			if(!from_string<int>(ato, s.substr(1), std::hex))
			{
				cerr<<"Number of steps inserted is not a number"<<endl;
				return retrieveInfo(is, m, s, mappedIndices);
			}
		}
		else if (s.substr(0,1) == "m")
		{
			if (macroEnabled)
			{/*
				// return the reply to the interactive refinement which will have to replace the refinement in the environment to the macro and at the end change back to itself

				MacroStepRefinement *macroR = dynamic_cast<MacroStepRefinement *>(env->getRefinement());

				// replace the refinement in the environment and continue processing, at the end, replace back to the previous refinement.

				// read the list of clauses indices in parentheses
				vector<int> vec(1);

				// read the target clause (second argument)
				Clause *c = new Clause();

				return new UserReply(UserReply::MACRO, vec, c);*/
			}
			else
			{
				cerr<<"Macro not supported by this refinement"<<endl;
				return retrieveInfo(is, m, s, mappedIndices);
			}
		}
		else
		{
			if (s.substr(0,1) != "s")
			{
				cerr<<"Character inserted is not t (Automatic) or s (Search)!"<<endl;
				return retrieveInfo(is, m, s, mappedIndices);
			}
			else
			{
				searchTerm = s.substr(1);
				if (s.size() < 2)
				{
					cerr<<"Search term must not be empty!"<<endl;
					return retrieveInfo(is, m, s, mappedIndices);
				}
			}
		}
	}
	else
	{
		cin>>b;
	}

	// automatic mode
	if (ato >= 0)
	{
		vector<int> vec(1);
		vec[0] = ato;
		return new UserReply(UserReply::AUTO, vec);
	}
	// normal mode
	else if (a > -1)
	{
		// TODO to confirm input values are in range and legal
		vector<int> vec(2);
		// q and b are the positions, extract the indices
		vec[0] = mappedIndices[a];
		vec[1] = mappedIndices[b];
		return new UserReply(UserReply::INTERACTIVE, vec);
	}
	// search mode
	else if (searchTerm.size() > 0)
	{
		set<Clause*> res = is->searchForClauses(searchTerm);
		for (set<Clause*>::const_iterator it2 = res.begin(); it2 != res.end(); ++it2)
		{
			Clause *cls = *it2;
			cout<<"("<<indexToLocation(m[cls], mappedIndices)<<") "<<cls->toString()<<"\t";
		}

		return retrieveInfo(is, m, s, mappedIndices);
	}
	// macro mode is being dealt with earlier in the method

	else
	{
		cerr<<"Unknown input command, please try again."<<endl;
		return retrieveInfo(is, m, s, mappedIndices);
	}
}

int PromptUserCommand::indexToLocation(int index, vector<int> &mappedIndices)
{
	for (int i=0; i<mappedIndices.size(); i++)
	{
		if (mappedIndices[i] == index)
		{
			return i;
		}
	}
	throw new ATPException("Unexpected behaviour in PromptUserCommand indexToLocation. Index is not in mappedIndices vector!");
}

UserReply *PromptUserCommand::getReplyFromUser(InputSet *is, map<Clause*, int> m, string s)
{
	// display map and create the Vector
	// the vector map the positions to indices (keys in the map)
	// so the indices can be extracted from the positions later on in the d=function

	vector<int> ind = printInfo(m);

	return retrieveInfo(is, m, s, ind);
}

void PromptUserCommand::errorEncountered(string msg)
{
	cerr<<msg<<endl;
}

void PromptUserCommand::clausesEventPerformed(InputSetEvent &r)
{
	switch (r.getType())
	{
	case InputSetEvent::CLAUSE_ADDED:
		cout<<r.getMessage()<<" with Unifier "<<r.getClause()->getClauseContext()->toString()<<endl;
		break;
	case InputSetEvent::CLAUSE_DELETED:
		cout<<r.getMessage()<<endl;
		break;
	case InputSetEvent::CLAUSE_REFUSED:
		cout<<r.getMessage()<<endl;
		break;
	}
}
