%skeleton "lalr1.cc" /* -*- C++ -*- */
%require "2.1a"
%defines
%define "parser_class_name" "ClausesParser"

%{
#include "Term.h"
#include "Constant.h"
#include "Function.h"
#include "Variable.h"
#include "datatypes.h"
#include "Clause.h"
#include "ConcreteClause.h"
#include "VariableNameGenerator.h"
#include "ClauseContext.h"

#include <string>
#include <vector>
#include <set>
class Prover9Driver;
%}

// The parsing context.
%parse-param { Prover9Driver& driver }
%lex-param { Prover9Driver& driver }

%locations
%initial-action
{
// Initialize the initial location.
@$.begin.filename = @$.end.filename = &driver.file;
};

%debug
%error-verbose

// Symbols.
%union
{
NS_PROOFLIB::Term *term;
NS_PROOFLIB::TermList *term_list;
std::string *sval;
Clause *clause;
std::vector<NS_PROOFLIB::Term*> *term_vector;
std::pair<std::vector<NS_PROOFLIB::Term*>*, std::vector<NS_PROOFLIB::Term*>* > *literal_vector;
std::set<Clause*> *clause_set;
};

%{
# include "Prover9Driver.h"
%}

%token END 0 "end of file"
%token OBRACE "("
%token EBRACE ")"
%token OCBRACE "{"
%token ECBRACE "}"
%token COMMA ","
%token <sval> CHAR "character"
%token <sval> CONST "constant character"
%token <sval> COMMENT "comment"
%token DOT "."
%token IMPLIED ":-"
%token EQ "="
%token QT "\""
%token NEG "-"
%token OR "|"

%type <term> term "term"
%type <term> function "function"
%type <term> constant "constant"
%type <term> variable "variable"
%type <term_list> term_list "term_list"
%type <clause> clause "clause"
%type <literal_vector> literal_vector "literal_vector"
%type <clause_set> clause_set "clause_set"

%printer { debug_stream () << *$$; } "character"
%printer { debug_stream () << $$->getStringRep(); } "function"
%printer { debug_stream () << $$->getStringRep(); } "constant"
%printer { debug_stream () << $$->getStringRep(); } "variable"
%printer { debug_stream () << $$->getStringRep(); } "term"

%destructor { delete $$; } "clause"
%destructor { delete $$->first; delete $$->second; delete $$; } "literal_vector"
%destructor { delete $$; } "clause_set"
%destructor { $$->detach(); } "function"
%destructor { $$->detach(); } "constant"
%destructor { $$->detach(); } "variable"

%destructor { $$->detach(); } "term"

%%

%start end_term;

end_term: clause_set { driver.result = $1; };

clause_set: clause_set clause
{
	$1->insert($2);
}
| /* Nothing */
{
	$$ = new std::set<Clause*>();
}
;

term: function {}
| constant {}
| variable {}
;

term_list: term_list COMMA term
{
	$1->addRight($3);
}
| term
{
	NS_PROOFLIB::TermList *l = new NS_PROOFLIB::TermList();
	l->addRight($1);
	$$ = l;
}
;

/* PROVER 9 */

clause: literal_vector DOT COMMENT
{
	Clause *cl = new ConcreteClause(*$1->first, *$1->second, driver.env->getVariableNameGenerator());
    cl->setClauseContext(new ClauseContext());
    $$ = cl;
}
| literal_vector DOT
{
	Clause *cl = new ConcreteClause(*$1->first, *$1->second, driver.env->getVariableNameGenerator());
    cl->setClauseContext(new ClauseContext());
    $$ = cl;
} 
;

literal_vector: literal_vector OR NEG term
{
	$1->second->push_back($4);
}
| literal_vector OR term
{
	$1->first->push_back($3);
}
| NEG term
{
	std::pair<std::vector<NS_PROOFLIB::Term*>*, std::vector<NS_PROOFLIB::Term*>* > *a = new std::pair<std::vector<NS_PROOFLIB::Term*>*, std::vector<NS_PROOFLIB::Term*>* >(new std::vector<NS_PROOFLIB::Term*>(),new std::vector<NS_PROOFLIB::Term*>());
	a->second->push_back($2);
	$$ = a;
}
| term
{
	std::pair<std::vector<NS_PROOFLIB::Term*>*, std::vector<NS_PROOFLIB::Term*>* > *a = new std::pair<std::vector<NS_PROOFLIB::Term*>*, std::vector<NS_PROOFLIB::Term*>* >(new std::vector<NS_PROOFLIB::Term*>(),new std::vector<NS_PROOFLIB::Term*>());
	a->first->push_back($1);
	$$ = a;
}
;

function: QT CHAR QT OBRACE term_list EBRACE
{
	$$ = driver.pdb->newFunction( $2->c_str(), *($5) ) ;
}
| CHAR OBRACE term_list EBRACE
{
	$$ = driver.pdb->newFunction( $1->c_str(), *($3) ) ;
}
| QT EQ QT OBRACE term_list EBRACE
{
	$$ = driver.pdb->newFunction( "=", *($5) ) ;
}
;


variable: CHAR {$$ = driver.pdb->newVariable( $1->c_str()) ;};

constant: QT CHAR QT {$$ = driver.pdb->newConstant( $2->c_str()) ;}
| CONST {$$ = driver.pdb->newConstant( $1->c_str()) ;}
;

%%

void
yy::ClausesParser::error (const yy::ClausesParser::location_type& l,
const std::string& m)
{
	driver.error (l, m);
}
