#include "ClauseFeatureVectorIndex.h"

using namespace std;
using namespace NS_PROOFLIB;

ClauseFeatureVectorIndex::ClauseFeatureVectorIndex()
{
	functionIndex = new map<string,unsigned int>();
}

ClauseFeatureVectorIndex::~ClauseFeatureVectorIndex()
{
	delete functionIndex;
}

void ClauseFeatureVectorIndex::init(std::set<Clause*>*st)
{
	set<string> symbols;
	// gather all function symbols in the input set
	typedef set<Clause *>::const_iterator SC;
	for (SC it = st->begin(); it != st->end(); ++it)
	{
		Clause *c = *it;
		vector<Term*> pos = c->getPositiveLiterals();
		vector<Term*> neg = c->getNegativeLiterals();
		typedef vector<Term*>::const_iterator VTI;
		for (VTI trm = pos.begin(); trm != pos.end(); ++trm)
		{
			addSymbols(symbols, *trm, c->getClauseContext());
		}
		for (VTI trm = neg.begin(); trm != neg.end(); ++trm)
		{
			addSymbols(symbols, *trm, c->getClauseContext());
		}
	}
	int i=0;
	for (set<string>::const_iterator it = symbols.begin(); it != symbols.end(); ++it)
	{
		(*functionIndex)[*it] = i++;
	}
	// for both pos and neg, pos are at functionIndex * 2 and neg are at
	// functionIndex * 2 + 1
}

void ClauseFeatureVectorIndex::addSymbols(set<string> &symbols, Term* t, ClauseContext *c)
{
	if (typeid(*t) == typeid(Variable))
	{
		t = c->getVariableInContext(t);
	}
	if (typeid(*t) == typeid(Constant))
	{
		symbols.insert(t->getSymbol());
	}
	else if (typeid(*t) == typeid(Function))
	{
		Function *f = dynamic_cast<Function *>(t);
		symbols.insert(t->getSymbol());			
		for (unsigned int i=1; i<=f->getArity(); i++)
		{
			addSymbols(symbols, f->getSub(i), c);
		}
	}
}

void ClauseFeatureVectorIndex::insert(Clause* c)
{
	ClauseFeatureVector v(functionIndex->size());
	v.init(c, functionIndex);
	addClauseToIndex(v,c);
}

void ClauseFeatureVectorIndex::addClauseToIndex(ClauseFeatureVector &v, Clause *c)
{
	// needs to travel branches according to the depth of the current symbol
	// symbols to depth should be already extracted by the vector v
	Trie *cur = &tree;
	// traverse path in inner nodes
	for (unsigned int i=0 ; i<functionIndex->size(); i++)
	{
		// pos symbol
		cur = cur->getOrCreate(v.getDepthAtIndex(i * 2));
		// neg symbol
		cur = cur->getOrCreate(v.getDepthAtIndex(i * 2 + 1));
	}
	// reached a leaf
	cur->insertClause(c);
}

string ClauseFeatureVectorIndex::toString()
{
	string ret;
	
	//Printing symbols and indices
	for (map<string,unsigned int>::const_iterator it = functionIndex->begin(); it != functionIndex->end(); ++it)
	{
		pair<string, unsigned int> p = *it;
		ret.append(p.first);
		ret.append(" - ");
		int i = p.second;
		std::string s;
		std::stringstream out;
		out << i;
		ret.append(out.str());
		ret.append("; ");
	}
	ret.append("\n");
	ret.append(tree.toString());
	return ret;
}

TrieIterator *ClauseFeatureVectorIndex::getPossibleSubsumers(Clause *c)
{
	ClauseFeatureVector *cfv = new ClauseFeatureVector(functionIndex->size());
	cfv->init(c, functionIndex);
	return new SubsumersIterator(cfv, &tree);
}

SubsumersIterator::SubsumersIterator(ClauseFeatureVector *cfv, Trie* t)
{
	vector = cfv;
	dfs.push(pair<Trie*,int>(t,0));
}

SubsumersIterator::~SubsumersIterator()
{
	delete vector;
}

Clause *SubsumersIterator::next()
{
	Clause *ret = NULL;
	
	if (!clauses.empty())
	{
		ret = clauses.top();
		clauses.pop();
	}
	else if (!dfs.empty())
	{
		pair<Trie*,unsigned int> p = dfs.top();
		dfs.pop();
		Trie *tree = p.first;
		set<Clause*> *cls = tree->clauses;
		
		// not a leaf node
		if (cls == NULL)
		{
			std::vector<Trie*> chld = *(tree->children);
			// get the depth of the new clause at the current index (p.second)
			unsigned int depth = vector->getDepthAtIndex(p.second);
			// insert all nodes which are smaller or equal in depth at index
			// TODO iterate backwards so minimal ones will be poped first - couldnt do it so far because of i-- does not work with unsigned int
			// make sure depth is not larger than number of children
			// (in whihc case take all children)
			unsigned int maxSize = depth;
			if (depth >= chld.size()) maxSize = chld.size() - 1;
			for (unsigned int i = 0; i <= maxSize; i++)
			{
				if (chld.at(i) != Trie::dummy)
				{
					dfs.push(pair<Trie*,unsigned int>(chld.at(i),p.second+1));
				}
			}
			return next();
		}
		// a leaf
		else
		{
			for (set<Clause*>::const_iterator it = cls->begin(); it != cls->end(); ++it)
			{
				clauses.push(*it);
			}
			ret = next();
		}
	}

	return ret;
}